/*
 *  IBM Confidential OCO Source Materials
 *
 *  5725-G92 Copyright IBM Corp. 2011, 2012
 *
 *  The source code for this program is not published or otherwise
 *  divested of its trade secrets, irrespective of what has
 *  been deposited with the U.S. Copyright Office.
 *
 */

/*
 * Program that executes a given program with the group-id set to a
 * given gid, and with the umask set to 002 (so that freshly created files
 * have the specified gid and are group-writable).
 *
 * @author Bruno Haible
 */

#include <sys/types.h>
#include <grp.h>

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#define MAX_REPLACEMENT_SIZE 1024

#if defined GROUP_TO_SET && defined PROGRAM_TO_CALL
static char numeric_group_string[] = GROUP_TO_SET;
static char callee[] = PROGRAM_TO_CALL;
#else
static char numeric_group_string[MAX_REPLACEMENT_SIZE] = "%GROUP_TO_SET%";
static char callee[MAX_REPLACEMENT_SIZE] = "%PROGRAM_TO_CALL%";
#endif

/* Tests whether the current user is a member of the given group. */
static int
is_member_of_group (gid_t gid)
{
  gid_t *grouplist;
  size_t grouplist_size;
  int ngroups;
  int i;

  if (getgid () == gid)
    return 1;

  grouplist_size = 1024;
  for (;;) {
    grouplist = (gid_t *) malloc (grouplist_size * sizeof (gid_t));
    if (grouplist == NULL) {
      fprintf (stderr, "gidwrapper: out of memory\n");
      exit (125);
    }
    ngroups = getgroups (grouplist_size, grouplist);
    if (ngroups >= 0)
      break;
    if (errno != EINVAL) {
      /* There's a problem fetching the group list. */
      free (grouplist);
      return 0;
    }
    free (grouplist);
    grouplist_size = 2 * grouplist_size;
  }
  for (i = 0; i < ngroups; i++) {
    if (grouplist[i] == gid) {
      free (grouplist);
      return 1;
    }
  }
  free (grouplist);
  return 0;
}

int
main (int argc, char *argv[])
{
  char *end;
  unsigned long gid = strtoul (numeric_group_string, &end, 10);
  if (end < numeric_group_string + strlen (numeric_group_string) || gid == ULONG_MAX) {
    fprintf (stderr, "gidwrapper: invalid numeric group id '%s'\n", numeric_group_string);
    exit (125);
  }

  /* This program is made setgid of the given gid. But we want to allow its execution only
     to those user ids which are member of this group - otherwise we have a security problem. */
  if (!is_member_of_group (gid)) {
    struct group *gr = getgrgid (gid);
    if (gr != NULL)
      fprintf (stderr, "This program can only be executed by users in group '%s'.\n", gr->gr_name);
    else
      fprintf (stderr, "This program can only be executed by users in group %u.\n", (unsigned int) gid);
    exit (126);
  }

  /* This program is made setgid of the given gid, therefore at this point the effective gid
     is the given one. But this is not enough for system calls such as access() and also
     not enough for Java - in these cases, the real gid still matters. So set the real gid
     as well. */
  if (setregid (gid, gid) < 0) {
    fprintf (stderr, "gidwrapper: setregid failed\n");
    exit (126);
  }

  umask (002);

  argv[0] = callee;
  return execv (callee, argv);
}

/*
 * For Emacs M-x compile.
 * Local Variables:
 * compile-command: "gcc -m32 -Wall -O -o ../../../../bin.linux-x86/gidwrapper.linux-x86.bin gidwrapper.c"
 * End:
 */
